--[[ Globals in use: ]]
-- gemsay_global: table of global categories and their strings (saved)
-- gemsay_char: table of per-char categories and their strings (saved)
-- gemsay_perchar, a Boolean: per-char or global categories?
-- gemsay_add, a Boolean: whether we're adding a new cat or renaming one
-- gemsay_ccat, a string: which category we're looking at ("" = none)
-- gemsay_cstr, an int: which string we're looking at (0 = none)

function GemSay_Cats_initialize()

  --[[ Initialize globals and such for the random message categories and their
       editor window ]]

  SlashCmdList["GSE"] = GemSay_gsr_command
  SLASH_GSE1 = "/gse"
  SlashCmdList["GSR"] = GemSay_gsr_command
  SLASH_GSR1 = "/gsr"
  if not gemsay_global then gemsay_global = {} end
  if not gemsay_char then gemsay_char = {} end
  gemsay_perchar = false
  gemsay_add = true
  gemsay_ccat = ""
  gemsay_cstr = 0
end

function GemSay_gsr_command()

  --[[ This command just pops up the category editor window. ]]

  getglobal("GemSay_Cats"):Show()
end

function GemSay_CatList_ScrollBar_update()

  --[[ Look at where the category list's scroll bar is and draw the category
       list accordingly. ]]

  local ncats
  local cats = {}
  local cat
  local line
  local offset = FauxScrollFrame_GetOffset(GemSay_Cats_CatList_ScrollBar)

  if gemsay_perchar then -- which table are we looking at?
    if gemsay_char then
      for cat in pairs(gemsay_char) do
        tinsert(cats, cat)
      end
    end
  else
    if gemsay_global then
      for cat in pairs(gemsay_global) do
        tinsert(cats, cat)
      end
    end
  end
  ncats = getn(cats)
  sort(cats)
  FauxScrollFrame_Update(GemSay_Cats_CatList_ScrollBar, ncats, 15, 16)
  for line = 1, 15 do
    local button = getglobal("GemSay_Cats_CatList_Button"..line)
    local lineplusoffset = line + offset

    if lineplusoffset <= ncats then
      getglobal("GemSay_Cats_CatList_Button"..line.."_Text"):SetText(cats[lineplusoffset])
      button.theName = cats[lineplusoffset]
      button:Show()
      if cats[lineplusoffset] == gemsay_ccat then -- highlight the selected one
        button:LockHighlight()
      else
        button:UnlockHighlight()
      end
    else
      button:Hide()
    end
  end
  if gemsay_ccat ~= "" then
    GemSay_Cats_CatList_ButtonFrame_RenameButton:Enable()
    GemSay_Cats_CatList_ButtonFrame_DeleteButton:Enable()
    GemSay_Cats_StringList_ButtonFrame_NewButton:Enable()
  else
    GemSay_Cats_CatList_ButtonFrame_RenameButton:Disable()
    GemSay_Cats_CatList_ButtonFrame_DeleteButton:Disable()
    GemSay_Cats_StringList_ButtonFrame_NewButton:Disable()
  end
end

function GemSay_StringList_ScrollBar_update()

  --[[ Look at where the string list's scroll bar is and draw the string
       list accordingly. ]]

  local nstrings
  local line
  local offset = FauxScrollFrame_GetOffset(GemSay_Cats_StringList_ScrollBar)

  if gemsay_perchar then -- which table?
    if (not gemsay_char) or (not gemsay_char[gemsay_ccat]) then
      nstrings = 0
    else
      nstrings = getn(gemsay_char[gemsay_ccat])
    end
  else
    if (not gemsay_global) or (not gemsay_global[gemsay_ccat]) then
      nstrings = 0
    else
      nstrings = getn(gemsay_global[gemsay_ccat])
    end
  end
  FauxScrollFrame_Update(GemSay_Cats_StringList_ScrollBar, nstrings, 15, 16)
  for line = 1, 15 do
    local button = getglobal("GemSay_Cats_StringList_Button"..line)
    local lineplusoffset = line + offset

    if lineplusoffset <= nstrings then
      if gemsay_perchar then
        getglobal("GemSay_Cats_StringList_Button"..line.."_Text"):SetText(gemsay_char[gemsay_ccat][lineplusoffset])
        button.theName = gemsay_char[gemsay_ccat][lineplusoffset]
      else
        getglobal("GemSay_Cats_StringList_Button"..line.."_Text"):SetText(gemsay_global[gemsay_ccat][lineplusoffset])
        button.theName = gemsay_global[gemsay_ccat][lineplusoffset]
      end
      button:Show()
      if lineplusoffset == gemsay_cstr then -- highlight the selected one
        button:LockHighlight()
      else
        button:UnlockHighlight()
    end
    else
      button:Hide()
    end
  end
  if gemsay_cstr ~= 0 then
    GemSay_Cats_StringList_ButtonFrame_EditButton:Enable()
    GemSay_Cats_StringList_ButtonFrame_DeleteButton:Enable()
  else
    GemSay_Cats_StringList_ButtonFrame_EditButton:Disable()
    GemSay_Cats_StringList_ButtonFrame_DeleteButton:Disable()
  end
end

function GemSay_ListSelect_changed(perchar)

  --[[ The user has clicked on either the global or per-character radio button.
       Deal with it.  In other words, change the state variable and redraw the
       lists, resetting the scroll bars. ]]

  gemsay_perchar = perchar
  gemsay_ccat = ""
  gemsay_cstr = 0
  GemSay_CatList_ScrollBar_update()
  GemSay_StringList_ScrollBar_update()
end

function GemSay_CatList_nameclick(button)

  --[[ A click should highlight the name of a category and activate the Rename
       and Delete buttons.  Clicking on a category button that was already
       highlighted should have the same result as clicking the Rename
       button. ]]

  local btext = button.theName

  if btext == "" then return end

  local again = (gemsay_ccat == btext)

  gemsay_ccat = btext
  if not again then
    gemsay_cstr = 0
  end
  GemSay_CatList_ScrollBar_update()
  GemSay_StringList_ScrollBar_update()
  if again then
    GemSay_rename_category()
  end
end

function GemSay_StringList_nameclick(button)

  --[[ A click should highlight a string and activate the Edit and Delete
       buttons.  Clicking on a string button that was already highlighted
       should have the same result as clicking the Edit button. ]]

  local bname = button:GetName()
  local bnum = tonumber(strsub(bname, 30))
  local offset = FauxScrollFrame_GetOffset(GemSay_Cats_StringList_ScrollBar)
  local cstr = bnum + offset
  local again = (gemsay_cstr == cstr)

  gemsay_cstr = cstr
  GemSay_CatList_ScrollBar_update()
  if again then
    GemSay_edit_string()
  end
end

function GemSay_new_category()

  --[[ Pressing this button should activate the category editing dialog and
       make the main editor window insensitive.  We should also remember that
       we're creating a new category. ]]

  gemsay_add = true
  GemSay_Cats.enableMouse = false
  GemSay_Cats.enableKeyboard = false
  GemSay_CatDialog_EditBox:SetText("")
  GemSay_CatDialog:Show()
end

function GemSay_rename_category()

  --[[ Pressing this button should activate the category editing dialog and
       make the main editor window insensitive.  We should also remember that
       we're editing an existing category name. ]]

  gemsay_add = false
  GemSay_Cats.enableMouse = false
  GemSay_Cats.enableKeyboard = false
  GemSay_CatDialog_EditBox:SetText(gemsay_ccat)
  GemSay_CatDialog:Show()
end

function GemSay_delete_category()

  --[[ Remove the selected category name from the category list and update it.
       Deactivate the Rename and Delete buttons if this reduces the number of
       categories in this list to zero. ]]

  local ncats

  if gemsay_perchar then
    gemsay_char[gemsay_ccat] = nil
    ncats = getn(gemsay_char)
  else
    gemsay_global[gemsay_ccat] = nil
    ncats = getn(gemsay_global)
  end
  gemsay_ccat = ""
  gemsay_cstr = 0
  GemSay_CatList_ScrollBar_update()
  GemSay_StringList_ScrollBar_update()
end

function GemSay_new_string()

  --[[ Pressing this button should activate the string editing dialog and
       make the main editor window insensitive.  We should also remember that
       we're creating a new string. ]]

  gemsay_add = true
  GemSay_Cats.enableMouse = false
  GemSay_Cats.enableKeyboard = false
  GemSay_StringDialog_EditBox:SetText("")
  GemSay_StringDialog:Show()
end

function GemSay_edit_string()

  --[[ Pressing this button should activate the string editing dialog and
       make the main editor window insensitive.  We should also remember that
       we're editing an existing string. ]]

  gemsay_add = false
  GemSay_Cats.enableMouse = false
  GemSay_Cats.enableKeyboard = false
  if gemsay_perchar then
    GemSay_StringDialog_EditBox:SetText(gemsay_char[gemsay_ccat][gemsay_cstr])
  else
    GemSay_StringDialog_EditBox:SetText(gemsay_global[gemsay_ccat][gemsay_cstr])
  end
  GemSay_StringDialog:Show()
end

function GemSay_delete_string()

  --[[ Remove the selected string from the string list and update it.
       Deactivate the Edit and Delete buttons if this reduces the number of
       strings in this list to zero. ]]

  if gemsay_perchar then
    tremove(gemsay_char[gemsay_ccat], gemsay_cstr)
  else
    tremove(gemsay_global[gemsay_ccat], gemsay_cstr)
  end
  gemsay_cstr = 0
  GemSay_StringList_ScrollBar_update()
end

function GemSay_CatDialog_OK_clicked()

  --[[ Look to see whether we were adding a new category or editing an old one.
       Make the necessary change to the necessary category list.  Update the
       list and activate the main editor window again.  And activate the Rename
       and Delete buttons if this was an add. ]]

  local newcat = GemSay_CatDialog_EditBox:GetText()

  -- Hitting OK on a blank box is fine, but does the same as Cancel.

  if newcat == "" then
    GemSay_Cats.enableMouse = true
    GemSay_Cats.enableKeyboard = true
    return
  end

  -- Don't let the user create a category named the same as an existing one.

  if (gemsay_perchar and gemsay_char[newcat])
     or (not gemsay_perchar and gemsay_global[newcat]) then
    message("You already have a category named '"..newcat.."'.")
    GemSay_CatDialog:Show()
    return
  end

  --[[ Don't let the user create a category with illegal characters in it
       (legal characters are letters, numbers, and _). ]]

  if strfind(newcat, "[^_%w]") then
    message("Only letters, numbers and _ are allowed in category names.")
    GemSay_CatDialog:Show()
    return
  end

  if gemsay_add then
    if gemsay_perchar then
      gemsay_char[newcat] = {}
    else
      gemsay_global[newcat] = {}
    end
  else
    if gemsay_perchar then
      gemsay_char[newcat] = gemsay_char[gemsay_ccat]
      gemsay_char[gemsay_ccat] = nil
    else
      gemsay_global[newcat] = gemsay_global[gemsay_ccat]
      gemsay_global[gemsay_ccat] = nil
    end
  end
  gemsay_ccat = newcat
  GemSay_CatList_ScrollBar_update()
  GemSay_StringList_ScrollBar_update()
  GemSay_Cats.enableMouse = true
  GemSay_Cats.enableKeyboard = true
end

function GemSay_CatDialog_Cancel_clicked()

  --[[ Activate the main editor window again. ]]

  GemSay_Cats.enableMouse = true
  GemSay_Cats.enableKeyboard = true
end

function GemSay_StringDialog_OK_clicked()

  --[[ Look to see whether we were adding a new string or editing an old one.
       Make the necessary change to the necessary string list.  Update the
       list and activate the main editor window again.  And activate the Edit
       and Delete buttons if this was an add. ]]

  local newstr = GemSay_StringDialog_EditBox:GetText()

  -- No blank strings.

  if newstr == "" then
    GemSay_Cats.enableMouse = true
    GemSay_Cats.enableKeyboard = true
    return
  end

  if gemsay_add then
    if gemsay_perchar then
      tinsert(gemsay_char[gemsay_ccat], newstr)
      sort(gemsay_char[gemsay_ccat])
    else
      tinsert(gemsay_global[gemsay_ccat], newstr)
      sort(gemsay_global[gemsay_ccat])
    end
  else
    if gemsay_perchar then
      gemsay_char[gemsay_ccat][gemsay_cstr] = newstr
      sort(gemsay_char[gemsay_ccat])
    else
      gemsay_global[gemsay_ccat][gemsay_cstr] = newstr
      sort(gemsay_global[gemsay_ccat])
    end
  end
  GemSay_CatList_ScrollBar_update()
  GemSay_StringList_ScrollBar_update()
  GemSay_Cats.enableMouse = true
  GemSay_Cats.enableKeyboard = true
end

function GemSay_StringDialog_Cancel_clicked()

  --[[ Activate the main editor window again. ]]

  GemSay_Cats.enableMouse = true
  GemSay_Cats.enableKeyboard = true
end


--[[ Spells: Spell buttons ]]

local function spellFunc(self, button)
   -- This one was a huge pain in the button!
   if IsShiftKeyDown() then
      local offset
      local page
      if SpellBookFrame.bookType == BOOKTYPE_PET then
	 offset = 0
	 page = SPELLBOOK_PAGENUMBERS[BOOKTYPE_PET]
      else
	 offset = SpellBookFrame.selectedSkillLineOffset
	 page = SPELLBOOK_PAGENUMBERS[SpellBookFrame.selectedSkillLine]
      end
      local sid = self:GetID() + offset + (SPELLS_PER_PAGE*(page - 1))
      local seb = GemSay_StringDialog_EditBox
      if seb:HasFocus() then
	 seb:Insert(GetSpellLink(sid, SpellBookFrame.bookType))
      end
   end
end

hooksecurefunc("SpellButton_OnModifiedClick", spellFunc)

local origContainerFunc = ContainerFrameItemButton_OnModifiedClick
function ContainerFrameItemButton_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_StringDialog_EditBox
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local pid = self:GetParent():GetID()
	 local link = GetContainerItemLink(pid, id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origContainerFunc(...)
end

local origBankFunc = BankFrameItemButtonGeneric_OnModifiedClick
function BankFrameItemButtonGeneric_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_StringDialog_EditBox
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local link = GetContainerItemLink(BANK_CONTAINER, id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origBankFunc(...)
end

local origDollFunc = PaperDollItemSlotButton_OnModifiedClick
function PaperDollItemSlotButton_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_StringDialog_EditBox
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local link = GetInventoryItemLink("player", id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origDollFunc(...)
end

local origInboxFunc = InboxFrame_OnModifiedClick
function InboxFrame_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_StringDialog_EditBox
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local link = GetInboxItemLink(InboxFrame.openMailID, id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origInboxFunc(...)
end

local origMerchantFunc = MerchantItemButton_OnModifiedClick
function MerchantItemButton_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_StringDialog_EditBox
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local link = GetMerchantItemLink(id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origMerchantFunc(...)
end

local origHyperLinkFunc = ChatFrame_OnHyperlinkShow
function ChatFrame_OnHyperlinkShow(...)
   local args = {...}
   local self = args[1]
   local linktext = args[3]
   if IsShiftKeyDown() then
      local editbox = GemSay_StringDialog_EditBox
      if editbox:HasFocus() then
	 editbox:Insert(linktext)
	 return nil
      end
   end
   return origHyperLinkFunc(...)
end
