--[[ A discussion of the gemsay_events data structure:

     The global gemsay_events (saved per character) is an array whose keys are
     mnemonic strings composed by this program, and whose values are arrays
     with more keys and values.  For example, gemsay_events["Cast:
     Polymorph"].event = "UNIT_SPELLCAST_SENT", .detail = "Polymorph", .target
     = "hostile", .dest = "rps", .msg = "You're about to be a sheep, %t!".
     The main key for each entry is just composed when the user creates or
     edits the entry: they will drag or drop a spell/ability or item icon onto
     the box the editor provides, then we'll see whether it's a spell/ability
     or an item, and if it's an item we'll see what the name of the spell it
     casts is.  The user provides the command.

]]

local lasttarget = ""
local lastspell = ""
local lastrank = ""
local capturing = false

function GemSay_Event_initialize()

  --[[ Initialize things about the event listening code and its GUI window. ]]

  SlashCmdList["GST"] = GemSay_gst_command
  SLASH_GST1 = "/gst"
  if not gemsay_events then gemsay_events = {} end
  gemsay_handled = {
    PLAYER_ENTER_COMBAT = "Attack",
    PET_ATTACK_START = "Pet Attack",
    PLAYER_DEAD = "Death",
    PLAYER_LEVEL_UP = "Level Up",
    UNIT_PET_EXPERIENCE = "Pet Level Up",
    UNIT_PET_TRAINING_POINTS = "Pet Training Points",
    PLAYER_LOGIN = "Login",
    PLAYER_LOGOUT = "Logout",
    UNIT_SPELLCAST_SENT = "Cast Button Pressed",
    UNIT_SPELLCAST_FAILED = "Cast Failed",
    UNIT_SPELLCAST_START = "Cast Started",
    UNIT_SPELLCAST_DELAYED = "Cast Delayed",
    UNIT_SPELLCAST_INTERRUPTED = "Cast Interrupted",
    UNIT_SPELLCAST_STOP = "Cast Stopped",
    UNIT_SPELLCAST_SUCCEEDED = "Cast Succeeded",
    UNIT_SPELLCAST_CHANNEL_START = "Channel Start",
    UNIT_SPELLCAST_CHANNEL_STOP = "Channel Stop",
    GEMSAY_SPELL_ABSORBED = "Miss: Absorbed",
    GEMSAY_SPELL_BLOCKED = "Miss: Blocked",
    GEMSAY_SPELL_DEFLECTED = "Miss: Deflected",
    GEMSAY_SPELL_DODGED = "Miss: Dodged",
    GEMSAY_SPELL_EVADED = "Miss: Evaded",
    GEMSAY_SPELL_IMMUNE = "Miss: Immune",
    GEMSAY_SPELL_MISSED = "Miss: Missed",
    GEMSAY_SPELL_PARRIED = "Miss: Parried",
    GEMSAY_SPELL_REFLECTED = "Miss: Reflected",
    GEMSAY_SPELL_RESISTED = "Miss: Resisted",
  }
  gemsay_description = {
    PLAYER_ENTER_COMBAT = "When you enter melee combat.",
    PET_ATTACK_START = "When your pet attacks.",
    PLAYER_DEAD = "When you die.",
    PLAYER_LEVEL_UP = "When you level up.",
    UNIT_PET_EXPERIENCE = "When your pet levels up.",
    UNIT_PET_TRAINING_POINTS = "When your pet earns Training Points.",
    PLAYER_LOGIN = "When you log in.",
    PLAYER_LOGOUT = "When you log out.",
    UNIT_SPELLCAST_SENT = "When you press a button to cast a spell.",
    UNIT_SPELLCAST_FAILED = "When your cast fails (LoS, target dead, etc.).",
    UNIT_SPELLCAST_START = "When you begin to cast a spell.",
    UNIT_SPELLCAST_DELAYED = "When your casting is delayed.",
    UNIT_SPELLCAST_INTERRUPTED = "When your cast is interrupted.",
    UNIT_SPELLCAST_STOP = "When you stop casting a spell.",
    UNIT_SPELLCAST_SUCCEEDED = "When you cast a spell successfully.",
    UNIT_SPELLCAST_CHANNEL_START = "When you start channeling a spell.",
    UNIT_SPELLCAST_CHANNEL_STOP = "When you stop channeling a spell.",
    GEMSAY_SPELL_ABSORBED = "When your target absorbs all your attack's damage.",
    GEMSAY_SPELL_BLOCKED = "When your target blocks your attack.",
    GEMSAY_SPELL_DEFLECTED = "When your target deflects your attack.",
    GEMSAY_SPELL_DODGED = "When your target dodges your attack.",
    GEMSAY_SPELL_EVADED = "When your ability's target evades.",
    GEMSAY_SPELL_IMMUNE = "When your spell's target is immune to your spell.",
    GEMSAY_SPELL_MISSED = "When your ability or attack misses.",
    GEMSAY_SPELL_PARRIED = "When your target parries your attack.",
    GEMSAY_SPELL_REFLECTED = "When your target reflects your spell back.",
    GEMSAY_SPELL_RESISTED = "When your spell's target resists.",
  }
  for event in pairs(gemsay_description) do
     _G["GemSay_Event_Edit_Editing_"..event.."_Label"]:SetText(gemsay_description[event])
  end
  gemsay_chand = ""
  gemsay_editing_event = ""
  UIDropDownMenu_EnableDropDown(GemSay_Event_Edit_Editing_Menu)
--  UIDropDownMenu_EnableDropDown(GemSay_Event_Edit_Editing_Command_Menu)
end

function GemSay_gst_command()

  --[[ This command just pops up the event handler editor window. ]]

   _G["GemSay_Event"]:Show()
end

function GemSay_Event_stringify_table(thetable)
  local stringtable = {}
  for i, val in pairs(thetable) do
    if(type(val) == "number") then
      toinsert = tostring(val)
    elseif(type(val) == "boolean") then
      if(val) then
        toinsert = "true"
      else
        toinsert = "false"
      end
    elseif(val == nil) then
      toinsert = "nil"
    else
      toinsert = val
    end
    tinsert(stringtable, toinsert)
  end
  return stringtable
end

function GemSay_Event_print(event, args)

  --[[ Just prints info about an event. ]]

  GemSay_print(event.."("..table.concat(GemSay_Event_stringify_table(args), ", ")..")")
end

function GemSay_CombatLog_Event_print(event, args)
   local msg = event
   local stringargs = GemSay_Event_stringify_table(args)

   msg = msg.."("..stringargs[2]..", "..stringargs[5]..", "..stringargs[9]
   if strsub(stringargs[2], 1, 6) == "SPELL_" then
      msg = msg..", "..stringargs[13]
   end
   msg = msg..")"
   GemSay_print(msg)
end

function GemSay_EventList_ScrollBar_update()

  --[[ Look at where the event list's scroll bar is and draw the event
       list accordingly. ]]

  local nevs
  local evs = {}
  local ev
  local line
  local offset = FauxScrollFrame_GetOffset(GemSay_Event_EventList_ScrollBar)

  for ev in pairs(gemsay_events) do
    tinsert(evs, ev)
  end
  nevs = getn(evs)
  sort(evs)
  FauxScrollFrame_Update(GemSay_Event_EventList_ScrollBar, nevs, 15, 16)
  for line = 1, 15 do
     local button = _G["GemSay_Event_EventList_Button"..line]
     local lineplusoffset = line + offset

     if lineplusoffset <= nevs then
	_G["GemSay_Event_EventList_Button"..line.."_Text"]:SetText(evs[lineplusoffset])
	button.theName = evs[lineplusoffset]
	button:Show()
	if evs[lineplusoffset] == gemsay_chand then -- highlight the selected one
	   button:LockHighlight()
	else
	   button:UnlockHighlight()
	end
     else
	button:Hide()
     end
  end
  if gemsay_chand == "" then
     GemSay_Event_EventList_ButtonFrame_DeleteButton:Disable()
  else
    GemSay_Event_EventList_ButtonFrame_DeleteButton:Enable()
  end
end

function GemSay_Event_EditFrame_update()

  --[[ Fill in the right-hand edit frame with what needs to be there --
       including nothing, if that's the case. ]]

  if gemsay_chand == "" then
    GemSay_Event_Edit_Editing:Hide()
    return
  end
  for event in pairs(gemsay_handled) do
     local object = _G["GemSay_Event_Edit_Editing_"..event]

    if object == nil then
      GemSay_print("GemSay_Event_Edit_Editing_"..event.." is nil!")
    else
      object:Hide()
    end
  end
  GemSay_Event_Edit_Editing:Show()

  if gemsay_editing_event == nil or gemsay_editing_event == "" then
    UIDropDownMenu_SetSelectedValue(GemSay_Event_Edit_Editing_Menu, nil)
    UIDropDownMenu_SetText(GemSay_Event_Edit_Editing_Menu, "")
  else
    UIDropDownMenu_SetSelectedValue(GemSay_Event_Edit_Editing_Menu,
				    gemsay_editing_event)
    UIDropDownMenu_SetText(GemSay_Event_Edit_Editing_Menu,
			   gemsay_handled[gemsay_editing_event])

    local eventframe = "GemSay_Event_Edit_Editing_"..gemsay_editing_event
    local name = _G[eventframe.."_Spell_Name"]
    local icon = _G[eventframe.."_Spell_ActionButtonIcon"]

    if gemsay_editing_detail ~= nil and gemsay_editing_detail ~= "" then
      if name then
        name:SetText(gemsay_editing_detail)
      end
    else
      if name then
        name:SetText("")
      end
    end
    if gemsay_editing_texture ~= nil and gemsay_editing_texture ~= "" then
      if icon then
        icon:SetTexture(gemsay_editing_texture)
      end
    else
      if icon then
        icon:SetTexture("")
      end
    end
    _G[eventframe]:Show()
  end

  GemSay_Event_slider_set(GemSay_Event_Edit_Editing_Freq,
			  gemsay_editing_freq)
  GemSay_Event_Edit_Editing_Command_Dest:SetText(gemsay_editing_dest)
  GemSay_Event_Edit_Editing_Command_Msg:SetText(gemsay_editing_msg)
end

function GemSay_Event_icon_receivedrag(self)

  --[[ If the user drags an icon into the actionbutton, get the name of the
       spell associated with that icon (if any) and save it.  If there's no
       spell, say so. ]]

  local type, arg1, arg2 = GetCursorInfo()

  --[[ "item": itemID, link
       "spell": slot, "spell"/"pet"
       "macro": index
       "money": amount ]]

  if type == nil then type = "nil" end
  if arg1 == nil then arg1 = "nil" end
  if arg2 == nil then arg2 = "nil" end
  if type == "item" then
    local item, x, x, x, x, x, x, x, x, texture = GetItemInfo(arg1)
    local spell = GetItemSpell(arg1)

    if spell == nil then
      GemSay_error("This item does not cast a spell.")
      ClearCursor()
      return
    end
    if spell == nil then spell = "nil" end
    _G[self:GetName().."Icon"]:SetTexture(texture)
    _G[self:GetParent():GetName().."_Name"]:SetText(spell.."|n(from "..item..")")
    --GameTooltip:SetText(spell)
    ClearCursor()
    gemsay_editing_detail = spell
    gemsay_editing_texture = texture
  elseif type == "spell" then
    local spell, rank = GetSpellBookItemName(arg1, arg2)
    local texture = GetSpellTexture(arg1, arg2)

    if spell == "Polymorph" then
      --[[ Polymorph(Pig) actually has rank equal to "Pig". ]]
      if strsub(rank, 1, 4) ~= "Rank" then
	spell = spell.."("..rank..")"
      end
    end
    _G[self:GetName().."Icon"]:SetTexture(texture)
    _G[self:GetParent():GetName().."_Name"]:SetText(spell)
    --GameTooltip:SetText(spell)
    ClearCursor()
    gemsay_editing_detail = spell
    gemsay_editing_texture = texture
  else
    GemSay_error("This is not an item or spell.")
  end
end

function GemSay_Event_begin_capture()

  --[[ When the user has clicked the "Capture Next Cast" button while editing a
       trigger, we come here to set the global variable indicating we're
       capturing events. ]]

  GemSay_print("Capture!  The next spell/ability/item you activate will appear here.")
  capturing = true
end

function GemSay_Event_end_capture(event, args)

  --[[ The user has activated a spell or item after clicking "Capture Next
       Cast."  Handle it. ]]

  local spellname, spellrank, spellicon

  if not capturing then return end
  if event == nil or event == "" then
    GemSay_print("Capture canceled.")
    capturing = false
    return
  end
  spellname = args[2]
  x, x, spellicon = GetSpellInfo(spellname)
  if spellicon == nil then spellicon = "" end
  gemsay_editing_detail = spellname
  gemsay_editing_texture = spellicon

  local eventframe = "GemSay_Event_Edit_Editing_"..gemsay_editing_event

  if gemsay_editing_detail ~= nil and gemsay_editing_detail ~= "" then
     _G[eventframe.."_Spell_Name"]:SetText(gemsay_editing_detail)
  end
  if gemsay_editing_texture ~= nil and gemsay_editing_texture ~= "" then
     _G[eventframe.."_Spell_ActionButtonIcon"]:SetTexture(gemsay_editing_texture)
  else
     _G[eventframe.."_Spell_ActionButtonIcon"]:SetTexture(gemsay_editing_texture)
  end
  GemSay_print("Capture done.")
  capturing = false
end

function GemSay_EventList_nameclick(button)

  --[[ A click should highlight the name of an event handler, activate
       the Delete button, and show the edit frame. ]]

  local btext = button.theName

  if btext == "" then return end

  gemsay_chand = btext
  GemSay_EventList_ScrollBar_update()
  gemsay_editing_event = gemsay_events[gemsay_chand].event or ""
  gemsay_editing_freq = gemsay_events[gemsay_chand].freq or 100
--  gemsay_editing_command = gemsay_events[gemsay_chand].command or ""
  gemsay_editing_dest = gemsay_events[gemsay_chand].dest or ""
  gemsay_editing_msg = gemsay_events[gemsay_chand].msg or ""

  --[[ For the ones that have details, get the details ]]

  if strsub(gemsay_editing_event, 1, 13) == "GEMSAY_SPELL_"
   or strsub(gemsay_editing_event, 1, 15) == "UNIT_SPELLCAST_" then
    gemsay_editing_detail = gemsay_events[gemsay_chand].detail or ""
    gemsay_editing_texture = gemsay_events[gemsay_chand].texture or ""
  else
    gemsay_editing_detail = nil
    gemsay_editing_texture = nil
  end
  GemSay_Event_EditFrame_update()
end

function GemSay_Event_new()

  --[[ The user has clicked the New Trigger button, meaning we activate the
       right-hand frame and give it some blank values. ]]

  gemsay_chand = "** NEW **"
  gemsay_editing_event = ""
  gemsay_editing_detail = ""
  gemsay_editing_texture = ""
  gemsay_editing_freq = 100
--  gemsay_editing_command = ""
  gemsay_editing_dest = ""
  gemsay_editing_msg = ""
  GemSay_Event_EditFrame_update()
end

function GemSay_Event_delete()

  --[[ The user has clicked the Delete Trigger button, meaning we erase the
       selected trigger and update everything. ]]

  gemsay_events[gemsay_chand] = nil
  gemsay_chand = ""
  GemSay_EventList_ScrollBar_update()
  GemSay_Event_EditFrame_update()
end

function GemSay_Event_evmenu_init()

  --[[ This puts items into the event type dropdown menu. ]]

  if gemsay_handled then
    local reverse_events = {}
    local names = {}

    for event, name in pairs(gemsay_handled) do
      reverse_events[name] = event
      tinsert(names, gemsay_handled[event])
    end
    sort(names)

    for index, name in ipairs(names) do
      local event = reverse_events[name]
      local info = UIDropDownMenu_CreateInfo()

      info.text = name
      info.value = event
      info.func = GemSay_Event_evmenu_onclick
      info.checked = false

      UIDropDownMenu_AddButton(info)
    end
  end
end

function GemSay_Event_evmenu_onload(self)

  --[[ This sets up the event type dropdown menu. ]]

  UIDropDownMenu_Initialize(self, GemSay_Event_evmenu_init)
  UIDropDownMenu_SetWidth(self, 200);
  UIDropDownMenu_JustifyText(self, "LEFT");
end

function GemSay_Event_evmenu_onclick(self, arg1, arg2, checked)

  --[[ The user has selected something from the event type dropdown menu.  Show
       the proper editing frame for this type of event. ]]

  gemsay_editing_event = self.value
  UIDropDownMenu_SetSelectedValue(GemSay_Event_Edit_Editing_Menu,
				  gemsay_editing_event)
  --GemSay_Event_Edit_Editing_Menu:SetText(gemsay_handled[gemsay_editing_event])
  GemSay_Event_EditFrame_update()
end

function GemSay_Event_cancel()

  --[[ The user has pressed the Cancel button.  Forget we were editing anything
       and hide the right-hand frame. ]]

  GemSay_Event_end_capture()
  gemsay_editing_event = ""
  gemsay_editing_detail = ""
  gemsay_editing_texture = ""
  gemsay_editing_freq = 100
  gemsay_editing_dest = ""
  gemsay_editing_msg = ""
  gemsay_chand = ""
  GemSay_Event_EditFrame_update()
  GemSay_EventList_ScrollBar_update()
  GemSay_Event_Edit_Editing:Hide()
end

function GemSay_Event_save()

  --[[ The user has pressed the Save button.  Check to make sure there's valid
       data wherever there has to be valid data, and if so, store the data and
       hide the right-hand frame. ]]

  GemSay_Event_end_capture()
  if gemsay_editing_event == "" then
    GemSay_error("You must specify an event.")
    return
  end
--  if gemsay_editing_command ~= "gs" and gemsay_editing_command ~= "gsw" then
--    GemSay_error("You must specify a command: either /gs or /gsw.")
--    return
--  end
  if GemSay_Event_Edit_Editing_Command_Dest:GetText() == "" then
    GemSay_error("You must specify a destination.")
    return
  end
  if GemSay_Event_Edit_Editing_Command_Msg:GetText() == "" then
    GemSay_error("You must enter a message.")
    return
  end
  if evname == "Cast Button Pressed"
    or evname == "Channel Start"
    or evname == "Channel Stop"
    or evname == "Cast Delayed"
    or evname == "Cast Failed"
    or evname == "Cast Interrupted"
    or evname == "Cast Resisted"
    or evname == "Cast Started"
    or evname == "Cast Stopped"
    or evname == "Cast Succeeded" then
    if gemsay_editing_detail == "" then
      GemSay_error("You must specify a spell or item.")
      return
    end
  end
  local evname = gemsay_handled[gemsay_editing_event]

  if evname == "Cast Button Pressed"
    or evname == "Channel Start"
    or evname == "Channel Stop"
    or evname == "Cast Delayed"
    or evname == "Cast Failed"
    or evname == "Cast Interrupted"
    or evname == "Cast Resisted"
    or evname == "Cast Started"
    or evname == "Cast Stopped"
    or evname == "Cast Succeeded" then
    gemsay_chand = evname..": "..gemsay_editing_detail
  else
    gemsay_chand = evname
  end
  if gemsay_events == nil then gemsay_events = {} end
  if gemsay_events[gemsay_chand] == nil then
    gemsay_events[gemsay_chand] = {}
  end
  gemsay_events[gemsay_chand].event = gemsay_editing_event
  gemsay_events[gemsay_chand].freq = GemSay_Event_Edit_Editing_Freq:GetValue()
--  gemsay_events[gemsay_chand].command = gemsay_editing_command
  gemsay_events[gemsay_chand].dest = GemSay_Event_Edit_Editing_Command_Dest:GetText()
  gemsay_events[gemsay_chand].msg = GemSay_Event_Edit_Editing_Command_Msg:GetText()
  if evname == "Cast Button Pressed"
    or evname == "Channel Start"
    or evname == "Channel Stop"
    or evname == "Cast Delayed"
    or evname == "Cast Failed"
    or evname == "Cast Interrupted"
    or evname == "Cast Resisted"
    or evname == "Cast Started"
    or evname == "Cast Stopped"
    or evname == "Cast Succeeded" then
    gemsay_events[gemsay_chand].detail = gemsay_editing_detail
    gemsay_events[gemsay_chand].texture = gemsay_editing_texture
  end
  gemsay_chand = ""
  GemSay_EventList_ScrollBar_update()
  GemSay_Event_Edit_Editing:Hide()
end

function GemSay_Event_handle(event, args)

  --[[ A game event has occurred that we're listening for.  Handle it. ]]

  if capturing then
    if event == "UNIT_SPELLCAST_SENT"
     or event == "UNIT_SPELLCAST_START"
     or event == "UNIT_SPELLCAST_CHANNEL_START"
     or event == "UNIT_SPELLCAST_FAILED" then -- Handle captured spell
      GemSay_Event_end_capture(event, args)
    end
  end

  if strsub(event, 1, 14) == "UNIT_SPELLCAST" then --[[ Handle UNIT_SPELLCAST* ]]
    local spell = args[2]
    local rank = args[3]

    if spell == "Polymorph" then --[[ Handle Polymorph(Pig), etc. ]]
      if strsub(rank, 1, 4) ~= "Rank" then
	spell = spell.."("..rank..")"
      end
    end

    --[[ As UNIT_SPELLCAST_SENT is the only one of the UNIT_SPELLCAST_*
         events that gives you the target, we need to react to it whether
         the user has a trigger defined for it or not, so we can remember
         the target and use it for subsequent events.  Remember the spell
         and rank too, to check and make sure that subsequent events refer
         to the same cast, as much as is possible.  Might want to check out the
         cast ID, to see whether that would work. ]]

    if event == "UNIT_SPELLCAST_SENT" then
      lastspell = spell
      lastrank = rank
      lasttarget = args[4]
      GemSay_Event_issue(gemsay_handled[event]..": "..spell, args)
    elseif event == "UNIT_SPELLCAST_CHANNEL_START" then

      --[[ Note that a CHANNEL_START can happen without a SENT in some cases,
           notably "Summoning Stone Effect", which happens when someone
	   clicks on a summoning stone.  If we have a CHANNEL_START, check
           whether it's the same spell as the last SENT, using lasttarget if
           so, but if not, still send the event along without lasttarget. ]]

      if spell == lastspell and rank == lastrank then
        args[4] = lasttarget
      end
      GemSay_Event_issue(gemsay_handled[event]..": "..spell, args)
    elseif event == "UNIT_SPELLCAST_CHANNEL_STOP"
     or event == "UNIT_SPELLCAST_DELAYED"
     or event == "UNIT_SPELLCAST_FAILED"
     or event == "UNIT_SPELLCAST_INTERRUPTED"
     or event == "UNIT_SPELLCAST_START"
     or event == "UNIT_SPELLCAST_STOP"
     or event == "UNIT_SPELLCAST_SUCCEEDED" then
      if spell == lastspell and rank == lastrank then
        args[4] = lasttarget
        GemSay_Event_issue(gemsay_handled[event]..": "..spell, args)
      end
    end

--[[ Sometimes events send arguments that don't work very well within GemSay,
     so we recast them as GemSay pseudo-events -- not real events, but the rest
     of GemSay can handle them as if they were real. ]]

--[[ We want to check COMBAT_LOG_EVENT_UNFILTERED and not just
     COMBAT_LOG_EVENT, because if the user has filtered out the events in
     question in their current combat log settings, GemSay won't get those
     events. ]]

--[[ All combat log events' first 11 parameters are: timestamp, eventtype,
     hideCaster, sourceGUID, sourceName, sourceFlags, sourceFlags2, destGUID,
     destName, destFlags, and destFlags2 ]]

--[[ This is complicated; you have to check the sourceFlags argument's lowest
     bit to see whether the source of the combat log message is in fact the
     player. ]]

  elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
    if bit.band(args[6], 0x00000001) == 0x00000001 then
      if args[2] == "SPELL_MISSED" then

      --[[ All SPELL_* events' next 3 arguments are spellID, spellName and
           spellSchool. ]]

      --[[ SPELL_MISSED then has missType and amountMissed. ]]

        if args[12] == lastspell then

        --[[ The way we have things set up, the GEMSAY_SPELL_* handlers
             expect args[4] to be the target and don't look at anything
             else. ]]

          args[4] = lasttarget
          if missType == "ABSORB" then
	    event = "GEMSAY_SPELL_ABSORBED"
          elseif missType == "BLOCK" then
	    event = "GEMSAY_SPELL_BLOCKED"
 	  elseif missType == "DEFLECT" then
	    event = "GEMSAY_SPELL_DEFLECTED"
 	  elseif missType == "DODGE" then
	    event = "GEMSAY_SPELL_DODGED"
 	  elseif missType == "EVADE" then
	    event = "GEMSAY_SPELL_EVADED"
 	  elseif missType == "IMMUNE" then
	    event = "GEMSAY_SPELL_IMMUNE"
 	  elseif missType == "MISS" then
	    event = "GEMSAY_SPELL_MISSED"
 	  elseif missType == "PARRY" then
	    event = "GEMSAY_SPELL_PARRIED"
 	  elseif missType == "REFLECT" then
	    event = "GEMSAY_SPELL_REFLECTED"
 	  elseif missType == "RESIST" then
            event = "GEMSAY_SPELL_RESISTED"
          end
          GemSay_Event_issue(gemsay_handled[event]..": "..args[12], args)
        end -- if right spell
      end -- SPELL_MISSED
    end -- the player's event

  --[[ If there are no details and the trigger fires whenever the
       event occurs at all, like PLAYER_DEAD, do this: ]]

  elseif gemsay_handled[event] then
    GemSay_Event_issue(gemsay_handled[event], args)
  else
    GemSay_Event_print(event, args)
  end
end

function GemSay_Event_issue(a, args)

  --[[ If the user has a trigger for the event in question, and if the random
       factors align, send the user's message. ]]

  if not gemsay_events[a] then return end
  if math.random(100) > gemsay_events[a].freq then return end
  local line = gemsay_events[a].dest.." "..gemsay_events[a].msg

--  if gemsay_events[a].command == "gs" then
    GemSay_gs_command(line, args)
--  elseif gemsay_events[a].command == "gsw" then
--    GemSay_gsw_command(line, args)
--  end
end

function GemSay_Event_slider_set(slider, value)

  --[[ Set the value of a slider. ]]

  slider:SetValue(value)
  GemSay_Event_slider_update(slider)
end

function GemSay_Event_slider_update(slider)

  --[[ Update a slider so the text matches the value. ]]

  local name = slider:GetName()

  _G[name.."Text"]:SetText(slider:GetValue().."%")
end

--[[ Spells: Spell buttons ]]

local function spellFunc(self, button)
   -- This one was a huge pain in the button!
   if IsShiftKeyDown() then
      local offset
      local page
      if SpellBookFrame.bookType == BOOKTYPE_PET then
	 offset = 0
	 page = SPELLBOOK_PAGENUMBERS[BOOKTYPE_PET]
      else
	 offset = SpellBookFrame.selectedSkillLineOffset
	 page = SPELLBOOK_PAGENUMBERS[SpellBookFrame.selectedSkillLine]
      end
      local sid = self:GetID() + offset + (SPELLS_PER_PAGE*(page - 1))
      local editbox = GemSay_Event_Edit_Editing_Command_Msg
      if editbox:HasFocus() then
	 editbox:Insert(GetSpellLink(sid, SpellBookFrame.bookType))
      end
   end
   return true
end

hooksecurefunc("SpellButton_OnModifiedClick", spellFunc)

local origContainerFunc = ContainerFrameItemButton_OnModifiedClick
function ContainerFrameItemButton_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_Event_Edit_Editing_Command_Msg
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local pid = self:GetParent():GetID()
	 local link = GetContainerItemLink(pid, id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origContainerFunc(...)
end

local origBankFunc = BankFrameItemButtonGeneric_OnModifiedClick
function BankFrameItemButtonGeneric_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_Event_Edit_Editing_Command_Msg
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local link = GetContainerItemLink(BANK_CONTAINER, id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origBankFunc(...)
end

local origDollFunc = PaperDollItemSlotButton_OnModifiedClick
function PaperDollItemSlotButton_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_Event_Edit_Editing_Command_Msg
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local link = GetInventoryItemLink("player", id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origDollFunc(...)
end

local origInboxFunc = InboxFrame_OnModifiedClick
function InboxFrame_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_Event_Edit_Editing_Command_Msg
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local link = GetInboxItemLink(InboxFrame.openMailID, id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origInboxFunc(...)
end

local origMerchantFunc = MerchantItemButton_OnModifiedClick
function MerchantItemButton_OnModifiedClick(...)
   local args = {...}
   local self = args[1]
   if IsShiftKeyDown() then
      local editbox = GemSay_Event_Edit_Editing_Command_Msg
      if editbox:HasFocus() then
	 local id = self:GetID()
	 local link = GetMerchantItemLink(id)
	 editbox:Insert(link)
	 return nil
      end
   end
   return origMerchantFunc(...)
end

local origHyperLinkFunc = ChatFrame_OnHyperlinkShow
function ChatFrame_OnHyperlinkShow(...)
   local args = {...}
   local self = args[1]
   local linktext = args[3]
   if IsShiftKeyDown() then
      local editbox = GemSay_Event_Edit_Editing_Command_Msg
      if editbox:HasFocus() then
	 editbox:Insert(linktext)
	 return nil
      end
   end
   return origHyperLinkFunc(...)
end
